#include <sys/soundcard.h>
#include <sys/ioctl.h>
#include <stdint.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include "alsa_pcm_api.h"
#define NFRAGS 32
#define ESD_MAX_WRITE_SIZE (21*4096)

extern unsigned int stereo;

void alsa_audio_close(int _snd_fd)
{
    ioctl(_snd_fd, SNDRV_PCM_IOCTL_HW_FREE);
    close(_snd_fd);
}

int alsa_audio_write_s16(int _snd_fd, int16_t *buf, uint32_t bufsize)
{
    struct snd_xferi xferi;
    int j;
    xferi.result = 0;
    xferi.buf = (char*) buf;
    xferi.frames = (stereo ? (bufsize >> 1) : bufsize);
    j = ioctl(_snd_fd, SNDRV_PCM_IOCTL_WRITEI_FRAMES, &xferi);
	if (j >= 0) {
		j = xferi.result;
	}
    if(j == -32) { //-EPIPE
        ioctl(_snd_fd, SNDRV_PCM_IOCTL_PREPARE);
    }
    return j;
}

int alsa_audio_open(const char *devpath, uint8_t channels, uint32_t samplerate)
{
    int ver, snd_fd = -1;
    if (!devpath) {
        devpath = "/dev/snd/pcmC0D0p";
    }
    snd_fd = open(devpath, O_WRONLY);
    if (snd_fd < 0)
        return -1;
    if (ioctl(snd_fd, SNDRV_PCM_IOCTL_PVERSION, &ver) < 0)
        return -1;
    // Initialize hwparams structure.
	alsa_init_hwparams(snd_fd, SND_PCM_ACCESS_RW_INTERLEAVED, &samplerate, 1024, channels);
    return snd_fd;
}

void oss_audio_close(int _snd_fd)
{
    ioctl(_snd_fd, SNDCTL_DSP_SYNC);
    close(_snd_fd);
}

int oss_audio_write_s16(int _snd_fd, int16_t *buf, uint32_t bufsize)
{
    return write(_snd_fd, buf, bufsize*sizeof(int16_t));
}

int oss_audio_open(const char *devpath, uint8_t channels, uint32_t oss_audio_rate)
{
    int afd = -1, tmp = 0;

    if (!devpath) {
        devpath = "/dev/dsp";
    }

    /* open the sound device */
    if ((afd = open(devpath, O_WRONLY, 0)) == -1)
    {   /* Opening device failed */
        return(-2);
    }

    /* set the sound driver audio format for playback */
    tmp = AFMT_S16_NE;
    if (ioctl(afd, SNDCTL_DSP_SETFMT, &tmp) == -1)
    {   /* Fatal error */
        goto err;
    }

    /* set the sound driver number of channels for playback */
    tmp = (channels-1);
    if (ioctl(afd, SNDCTL_DSP_STEREO, &tmp) == -1)
    {   /* Fatal error */
        goto err;
    }

    /* set the sound driver number playback rate */
    tmp = oss_audio_rate;
    if (ioctl(afd, SNDCTL_DSP_SPEED, &tmp) == -1)
    { /* Fatal error */
        goto err;
    }

    return afd;
err:
    close(afd);
    return(-1);
}

