/***************************************************************************
 *   Copyright (C) 2004 by David Banz                                      *
 *   neko@netcologne.de                                                    *
 *   GPL'ed                                                                *
 ***************************************************************************/

#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include "tfmxsong.h"
#include "player.h"
#include <sys/stat.h>
#include <fcntl.h>
#include "md5.h"

static __attribute__((always_inline)) uint16_t av_bswap16(uint16_t x)
{
    x= (x>>8) | (x<<8);
    return x;
}

/* external functions */
void TfmxInit();
void StartSong();
int try_to_makeblock(S16 *buf);
extern uint32_t bqueue;

/* internal functions */
int check_md5_and_headers(char *mfile);
void usage(char *x);
void dump_macro(int *a);
void dump_pattern(int x);
int load_tfmx(char *mfn, char *sfn);
void do_debug(void);

/* MD5 digests for TFMX songs that need special treatment... */
unsigned const char* md5GemxTitle="\xf4\xa1\xd6\x04\xc4\xdf\xb3\x0f\x91\x0f\xd8\xac\x4b\x96\xe3\x06";
unsigned const char* md5DfreakTitle="\x9c\xf1\x72\x34\xbc\xe1\x0d\x21\xe2\x90\x11\x72\xeb\x1a\x2a\xcc";
unsigned const char* md5OopsUpBroken="\x79\x41\x33\xe1\xf1\x13\x1b\x23\xc0\x9c\xd7\x29\xe8\xa6\x2a\x4e";
unsigned const char* md5OopsUp="\x9a\x19\x78\x84\xa8\x15\x5e\xdd\x02\x90\x23\x57\xfa\xf2\x4e\x4e";

unsigned const char*
md5Monkey="\xc9\x5a\xa4\xf4\x44\xe8\x9a\x3f\x61\x4d\xfd\xe4\x20\x29\x96\x2a";

/* this is the MDAT that causes a segfault and other errors on MacOS-X */
unsigned const char* md5WeirdZoutThm="\xb2\x7c\xa7\x9c\x14\x69\x63\x87\xc1\x9c\x01\xf6\x5e\x15\x3e\xff";

int weirdZoutThm=0;
int dangerFreakHack=0;
int oopsUpHack=0;
int monkeyHack=0;

/* do we have a single-file TFMX (mdat+smpl in one file) ? */
int singleFile=0;
/* are DOS extensions used? (.tfx/.sam) */
int dosExt=0;
/* header data for single-file TFMX */
uint nTFhd_offset=0;
uint nTFhd_mdatsize=0;
uint nTFhd_smplsize=0;

U32 outRate=44100;

struct Hdr hdr;
extern struct Hdb hdb[8];
extern struct Pdblk pdb;
extern int LoopOff();
extern struct Mdb mdb;

extern char act[8];

char outf[PATHNAME_LENGTH]="/dev/null";
unsigned int mlen;
U32 editbuf[65536];
U8 *smplbuf;
int *macros;
int *patterns;
short ts[512][8];

int num_ts,num_pat,num_mac;

int songnum=0;
int startPat=-1;
int gemx=0;
int loops=1;
extern int blend,filt;

/* misc vars for TFMX format test */
#define MAGIK_LEN 11        /* length including final null byte */
unsigned int nMagikLen=MAGIK_LEN;
char pMagikBuf[MAGIK_LEN];

int check_md5_and_headers(char *mfile)
{
	MD5_CTX mdctx;
	unsigned char md_value[MD5_DIGEST_STRING_LENGTH];
	int ooxc=0,ooxc2=0;
	long size=-1;
	unsigned char* fdat;

    int fd = -1;	
    struct stat st;

	/* allocale buffer */
    if (stat(mfile, &st) < 0) {
	    perror("fstat");
		return(1);
    }
    size = st.st_size;	
	fdat=malloc((unsigned long)size);
	/* fill buffer */
    if((fd = open(mfile, O_RDONLY)) < 0) {
        perror("open");
        return(1);
    }
	if (read(fd, fdat, st.st_size) < 0) {
	    perror("read");
		close(fd);
		return(1);
	}
    close(fd);

	/* since we have got the file loaded, it's a good time to run the format test... */
	/* first clear magik buffer */
	memcpy(pMagikBuf,"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",nMagikLen);

	/* if it's a single-file TFMX, let's grab the pointers to mdat+smpl while we're at it */
	if (singleFile==1)
	{
		memcpy((void*)(&nTFhd_offset), (void*)(fdat+4), 4);
		nTFhd_offset=av_bswap32(nTFhd_offset);
		/* now we have the size of the header */
		memcpy((void*)(&nTFhd_mdatsize), (void*)(fdat+10), 4);
		nTFhd_mdatsize=av_bswap32(nTFhd_mdatsize);
		/* now we have the size of the mdat */
		memcpy((void*)(&nTFhd_smplsize), (void*)(fdat+14), 4);
		nTFhd_smplsize=av_bswap32(nTFhd_smplsize);
		/* now we have the size of the smpl */
	
		/* check if the actual filesize matches the size given by the header */
		if (nTFhd_offset+nTFhd_mdatsize+nTFhd_smplsize != size)
		{
			printf("\nERROR! 1-file TFMX header defines illegal size:\
[%d] instead of [%ld] \n",nTFhd_offset+nTFhd_mdatsize+nTFhd_smplsize,(unsigned long)size);
			return(1);
		}
	
	}

	if (memcmp("TFMX",fdat,4)) {
		return(2);
	}

	/* create md5 digest */
	MD5Init(&mdctx);
	MD5Update(&mdctx, fdat, size);
	MD5Final(&mdctx, md_value);
	
	/* compare md5 sums */
    dangerFreakHack = !memcmp(md_value, md5DfreakTitle, 16);
    gemx = !memcmp(md_value, md5GemxTitle, 16);
    weirdZoutThm = !memcmp(md_value, md5WeirdZoutThm, 16);
    monkeyHack = !memcmp(md_value, md5Monkey, 16);
    ooxc  = !memcmp(md_value, md5OopsUp, 16);
    ooxc2 = !memcmp(md_value, md5OopsUpBroken, 16);
	if (ooxc || ooxc2)
	    oopsUpHack=1;

#ifdef WORDS_BIGENDIAN
	if (weirdZoutThm==1) {
		printf("Warning! Problematic Z-Out theme mdat detected!\n\
May cause crashes/hangups on big-endian CPUs!\n");
	}
#endif

	/* free buffer */
	free(fdat);
	return(0);
}

void usage(char *x)
{
	printf("tfmxplay v1.1.6 by Jon Pickard <marxmarv@antigates.com>,\n"
"Neochrome <neko@netcologne.de> and others.\n"
"Copyright 1996-2004, see accompanying README for details.\n"
"\n"
"Usage: %s [options] mdat-file [smpl-file]\n"
"where options is one or more of:\n"
"-b mode	set stereo mode (0=mono, default 1=headphone, 2=stereo)\n"
"-p num		subsong to play (default 0)\n"
"-f freq	suggest playback rate in samples/sec (default 44100)\n"
"-o file	write audio output to file\n"
"-w num		set low-pass filter frequency (0=none, 3=lowest, default 0)\n"
"-l num		set loop mode (0=no repeat, default 1=infinite)\n"
"-D         force hack for Danger Freak title tune\n"
"-G         force old hack for GemX title tune (still incomplete)\n"
,x
);

}

void dump_macro(int *a)
{
	int x=0,s=0;
	while (((x&0xff000000)!=(0x07000000))&&(s<511))
	{
		x=av_bswap32(a[s]);
		printf("%04x: %08x\n",s,x);
		s++;
	}
}

/* this one can also load a single-file TFMX :) */
int load_tfmx(char *mfn, char *sfn)
{
    int fd = -1;
	struct stat s;
	unsigned int x,y,z=0;
	U16 x0,*sh,*lg;

    if((fd = open(mfn, O_RDONLY)) < 0)
	{
		perror("open");
		return(1);
	}

	/* jump to mdat start if single-file format */
	if (singleFile==1)
	{
		lseek(fd, nTFhd_offset, SEEK_CUR);
	}

	if (read(fd, &hdr, sizeof(hdr)) <= 0)
	{
		perror("read");
        close(fd);
		return(1);
	}

	if (strncmp("TFMX-SONG",hdr.magic,9)&&
	    strncmp("TFMX_SONG",hdr.magic,9)&&
	    strncasecmp("TFMXSONG",hdr.magic,8) &&
	    strncmp("TFMX",hdr.magic,4))
	{
        close(fd);
		return(2);
	}

    if((x = read(fd, &editbuf, 16384*sizeof(int))) < 0)
	{
		perror("read");
        close(fd);
		return(1);
	}

	/* close file if we have two files for mdat+smpl */
	if (singleFile==0)
	{
        close(fd);
	}

    printf("x: %u\n", x);
	mlen=x;
	editbuf[x]=-1;

	if (!hdr.trackstart)
		hdr.trackstart=0x180;
	else
		hdr.trackstart=(av_bswap32(hdr.trackstart)-0x200)>>2;

	if (!hdr.pattstart)
		hdr.pattstart=0x80;
	else
		hdr.pattstart=(av_bswap32(hdr.pattstart)-0x200)>>2;

	if (!hdr.macrostart)
		hdr.macrostart=0x100;
	else
		hdr.macrostart=(av_bswap32(hdr.macrostart)-0x200)>>2;

	if (x<136)
	{
		return(2);
	}

	for (x=0;x<32;x++)
	{
		hdr.start[x]=av_bswap16(hdr.start[x]);
		hdr.end[x]=av_bswap16(hdr.end[x]);
		hdr.tempo[x]=av_bswap16(hdr.tempo[x]);
	}

/* Now that we have pointers to most everything, this would be a good time to
   fix everything we can... ntohs tracksteps, convert pointers to array
   indices, av_bswap32 patterns and macros.  We fix the macros first, then the
   patterns, and then the tracksteps (because we have to know when the
   patterns begin to know when the tracksteps end...) */
	z=hdr.macrostart;
	macros = &editbuf[z];

	for (x=0;x<128;x++)
	{
		y=(av_bswap32(editbuf[z])-0x200);
		if ((y&3)||((y>>2)>mlen)) /* probably not strictly right */
			break;
		editbuf[z++]=y>>2;
	}
	num_mac=x;

	z=hdr.pattstart;
	patterns = &editbuf[z];
	for (x=0;x<128;x++)
	{
		y=(av_bswap32(editbuf[z])-0x200);
		if ((y&3)||((y>>2)>mlen))
			break;
		editbuf[z++]=y>>2;
	}
	num_pat=x;

	lg=(U16 *)&editbuf[patterns[0]];
	sh=(U16 *)&editbuf[hdr.trackstart];
	num_ts=(patterns[0]-hdr.trackstart)>>2;
	y=0;
	while (sh<lg)
	{
		x0=*sh;
		*sh++=((x0>>8)|(x0<<8));
	}

/* Now at long last we load the sample file/data. */

	/* different handling for single- and dual-file formats */
	if (singleFile==1)
	{
	    /* jump to smpl start */
	    uint32_t nSmplPos=nTFhd_offset+nTFhd_mdatsize;
		lseek(fd, nSmplPos, SEEK_SET);
        /* allocate mem */
		if (!(smplbuf=(void *)malloc(nTFhd_smplsize))) {
            close(fd);
			return(1);
		}
		/* read samples */
		if (read(fd, smplbuf, nTFhd_smplsize) < 0)
		{
			perror("read");
            close(fd);
			free(smplbuf);
			return(1);
		}
		/* finally close the file */
        close(fd);
	} else {
		if ((y=open(sfn,O_RDONLY))<=0) {
			perror("open");
			return(1);
		}
		if (fstat(y,&s)) {
			perror("fstat");
			close(y);
			return(1);
		}
		smplbuf=(void *)malloc(s.st_size);
		if (read(y,smplbuf,s.st_size) <= 0) {
			perror("read");
			close(y);
			free(smplbuf);
			return(1);
		}
		close(y);
	}
	return (0);
}

char *pattcmds[]={
    (char *)"End --Next track  step--",
    (char *)"Loop[count     / step.w]",
    (char *)"Cont[patternno./ step.w]",
    (char *)"Wait[count 00-FF--------",
    (char *)"Stop--Stop this pattern-",
    (char *)"Kup^-Set key up/channel]",
    (char *)"Vibr[speed     / rate.b]",
    (char *)"Enve[speed /endvolume.b]",
    (char *)"GsPt[patternno./ step.w]",
    (char *)"RoPt-Return old pattern-",
    (char *)"Fade[speed /endvolume.b]",
    (char *)"PPat[patt./track+transp]",
    (char *)"Lock---------ch./time.b]",
    (char *)"----------No entry------",
    (char *)"Stop-Stop custompattern-",
    (char *)"NOP!-no operation-------"
};

char *macrocmds[]={
    (char *)"DMAoff+Resetxx/xx/xx flag/addset/vol   ",
    (char *)"DMAon (start sample at selected begin) ",
    (char *)"SetBegin    xxxxxx   sample-startadress",
    (char *)"SetLen      ..xxxx   sample-length     ",
    (char *)"Wait        ..xxxx   count (VBI''s)     ",
    (char *)"Loop        xx/xxxx  count/step        ",
    (char *)"Cont        xx/xxxx  macro-number/step ",
    (char *)"-------------STOP----------------------",
    (char *)"AddNote     xx/xxxx  note/detune       ",
    (char *)"SetNote     xx/xxxx  note/detune       ",
    (char *)"Reset   Vibrato-Portamento-Envelope    ",
    (char *)"Portamento  xx/../xx count/speed       ",
    (char *)"Vibrato     xx/../xx speed/intensity   ",
    (char *)"AddVolume   ....xx   volume 00-3F      ",
    (char *)"SetVolume   ....xx   volume 00-3F      ",
    (char *)"Envelope    xx/xx/xx speed/count/endvol",
    (char *)"Loop key up xx/xxxx  count/step        ",
    (char *)"AddBegin    xx/xxxx  count/add to start",
    (char *)"AddLen      ..xxxx   add to sample-len ",
    (char *)"DMAoff stop sample but no clear        ",
    (char *)"Wait key up ....xx   count (VBI''s)     ",
    (char *)"Go submacro xx/xxxx  macro-number/step ",
    (char *)"--------Return to old macro------------",
    (char *)"Setperiod   ..xxxx   DMA period        ",
    (char *)"Sampleloop  ..xxxx   relative adress   ",
    (char *)"-------Set one shot sample-------------",
    (char *)"Wait on DMA ..xxxx   count (Wavecycles)",
    (char *)"Random play xx/xx/xx macro/speed/mode  ",
    (char *)"Splitkey    xx/xxxx  key/macrostep     ",
    (char *)"Splitvolume xx/xxxx  volume/macrostep  ",
    (char *)"Addvol+note xx/fe/xx note/CONST./volume",
    (char *)"SetPrevNote xx/xxxx  note/detune       ",
    (char *)"Signal      xx/xxxx  signalnumber/value",
    (char *)"Play macro  xx/.x/xx macro/chan/detune ",
    (char *)"SID setbeg  xxxxxx   sample-startadress",
    (char *)"SID setlen  xx/xxxx  buflen/sourcelen  ",
    (char *)"SID op3 ofs xxxxxx   offset            ",
    (char *)"SID op3 frq xx/xxxx  speed/amplitude   ",
    (char *)"SID op2 ofs xxxxxx   offset            ",
    (char *)"SID op2 frq xx/xxxx  speed/amplitude   ",
    (char *)"SID op1     xx/xx/xx speed/amplitude/TC",
    (char *)"SID stop    xx....   flag (1=clear all)"
};

void dump_pattern(int x)
{
	const char *n1="CCDDEFFGGAAB",*n2=" # #  # # # ";
	UNI a;
	int y=0,z,zz;
	static char n[]={0,0,0,0};
	printf("Pattern %02x:\n",x);
	x=patterns[x];
	a.b.b0=0;
	while (a.b.b0!=0xF0)
	{
		a.l=av_bswap32(editbuf[x++]);
		if ((z=a.b.b0)<0xF0)
		{
			zz=(z&0x3F)+6;
			n[2]=48+(zz/12);
			zz%=12;
			n[0]=n1[zz];
			n[1]=n2[zz];
		}
		if (z<0x80)
		{
			printf("%04x: %02x %s %02x %x %x %02x\n",
			       y++,
			       a.b.b0,
			       n,
			       a.b.b1,
			       a.b.b2>>4,
			       a.b.b2&0xF,
			       a.b.b3
			       );
		}
		else if (z<0xC0)
		{
			printf("%04x: %02x %s %02x %x %x %02x wait\n",
			       y++,
			       a.b.b0,
			       n,
			       a.b.b1,
			       a.b.b2>>4,
			       a.b.b2&0xF,
			       a.b.b3
			       );
		}
		else if (z<0xF0)
		{
			printf("%04x: %02x %s %02x %x %x %02x porta\n",
			       y++,
			       a.b.b0,
			       n,
			       a.b.b1,
			       a.b.b2>>4,
			       a.b.b2&0xF,
			       a.b.b3
			       );
		}
		else
		{
			printf("%04x: %02x %s %02x %x %x %02x\n",
			       y++,
			       a.b.b0,
			       pattcmds[z-0xF0],
			       a.b.b1,
			       a.b.b2>>4,
			       a.b.b2&0xF,
			       a.b.b3
			       );
		}
	}
}

extern int alsa_audio_open(const char *devpath, uint8_t channels, uint32_t sample_rate);
extern int alsa_audio_write_s16(int _snd_fd, short *buf, unsigned int bufsize);
extern int oss_audio_open(const char *devpath, uint8_t channels, uint32_t sample_rate);
extern int oss_audio_write_s16(int _snd_fd, short *buf, unsigned int bufsize);

int main(int argc, char **argv)
{
	char* tfxloc=0;

	int x;
    uint32_t n;
	char* c=0;
	char *mfn=NULL,sfn[PATHNAME_LENGTH];
    short buf[65536];
    int toOutFile = 0;
    int pcm_fd = -1;

	filt = 0;
    setvbuf(stdout, NULL, _IONBF, 0);

	while ((x=getopt(argc,argv,"GDivSb:8o:f:P:V:p:w:l:"))!=-1)
	{
		switch (x)
		{
		case '?':
		case ':':
			usage(argv[0]);
			exit(2);
		case 'o':
			strncpy(outf,optarg,PATHNAME_LENGTH-1);
			outf[PATHNAME_LENGTH-1]='\0';
			toOutFile=1;
			break;
		case 'P':
			startPat=strtol(optarg,NULL,0);
			break;
		case 'f':
			outRate=strtol(optarg,NULL,0);
			break;
		case 'b':
			blend=strtol(optarg,NULL,0);
			break;
		case 'p':
			songnum=strtol(optarg,NULL,0);
			break;
		case 'w':
			filt=strtol(optarg,NULL,0);
			break;
		case 'l':
			loops=strtol(optarg,NULL,0);
			break;
		case 'v':
			break;
		case 'G':
			gemx=1;
			break;
		case 'D':
			dangerFreakHack=1;
			break;
		case 'S':
			/* left for compatibility, this switch is inactive */
			break;
		case 'V':
			c=optarg;
			for(;*c;act[(*c++)&7]=0);
			break;
		default:
			fprintf(stderr,"getopt: got code 0x%x\n",x);
			break;
		}
	}

	if (optind<argc)
	{
		mfn = argv[optind++];
        sfn[0]=0;

		if (optind<argc)
		{
			strncpy(sfn,argv[optind++],PATHNAME_LENGTH-1);
			sfn[PATHNAME_LENGTH-1]='\0';
			printf("IF\n");
		}
		else
		{
			strncpy(sfn,mfn,PATHNAME_LENGTH-1);
			sfn[PATHNAME_LENGTH-1]='\0';
			if (!(c=strrchr(sfn,'/'))) c=sfn; else c++;
			
			/* start looking for ".tfx" at the first null byte */
			tfxloc=strchr(c,'\0');

			/* filename w/o path must be at least 5 chars long */
			if ((tfxloc-4) > c)
			{
				/* move pointer to the last four chars */
				tfxloc=tfxloc-4;
				if (0 == strncasecmp(tfxloc,".tfx",4))
				{
					/* we are using dos extensions */
					dosExt=1;
				}
			}

			if (dosExt != 1)
			{
				if (strncasecmp(c,"mdat.",5))
				{
					if (strncasecmp(c,"tfmx.",5)) {
						puts("'mdat'/'tfmx' prefix missing\n");
					} else {
						singleFile=1;
						sfn[0]='\0';
					}
				}
	
				/*
				* Case-preserving conversion of "mdat" to "smpl"
				*/
				if (singleFile==0)
				{
					(*c++)^='m'^'s';
					(*c++)^='d'^'m';
					(*c++)^='a'^'p';
					(*c++)^='t'^'l';
					c-=4;
				}
			}
			else
			{
				/* we assume DOS extensions are used
				with dual-files only! */
				/*
				* Case-preserving conversion of ".tfx" to ".sam"
				*/
				*tfxloc++; /* skip the dot */
				(*tfxloc++)^='t'^'s';
				(*tfxloc++)^='f'^'a';
				(*tfxloc++)^='x'^'m';
				tfxloc-=4;
			}
		}
	}
	else
	{
		usage(argv[0]);
		return 2;
	}

/* ------- do all that MD5, magic and single-file header stuff ----------- */

	x=check_md5_and_headers(mfn);
	if (x && (x != 2)) {
		fprintf(stderr,"%s: check_md5_and_headers failed\n",argv[0]);
		return 1;
    } else if (x==2) {
		fprintf(stderr,"%s: Not an MDAT/TFMX file\n",c);
		return 2;
	}

/* ----------------------------------------------------------------------- */

	if ((x=load_tfmx(mfn,sfn))==1)
	{
		fprintf(stderr,"%s: load_tfmx failed\n",argv[0]);
		return 1;
	}
	else if (x==2)
	{
		fprintf(stderr,"%s: Not an MDAT/TFMX file\n",c);
		return 2;
	}

	if (blend) stereo=1;
	blend&=1;

	if (!(c=strrchr(mfn,'/'))) c=mfn; else c++;
	printf("Module: %s\n",c);

	for (x=0;x<6;x++) printf(">%40s\n",hdr.text[x]);
	printf("\n");

	printf("%d tracksteps at 0x%04x\n",num_ts,(hdr.trackstart<<2)+0x200);
	printf("%d patterns at 0x%04x\n",num_pat,(hdr.pattstart<<2)+0x200);
	printf("%d macros at 0x%04x\n",num_mac,(hdr.macrostart<<2)+0x200);

	for (x=0;x<31;x++) {
		if (/*(hdr.start[x]!=hdr.end[x]) &&*/ (hdr.end[x]))
			printf("Song %2d: start %3x end %3x\n",x,av_bswap16(hdr.start[x]),av_bswap16(hdr.end[x]));
    }

	/* FIXME debug output */
	if (monkeyHack == 1) {
		printf("MONKEY ISLAND DETECTED\n");
	}

/* Now the song is fully loaded.  Everything is done but av_bswap32'ing the actual
pattern and macro data. The routines that use the data do it for themselves.*/
    if(toOutFile) {
        if ((pcm_fd=open(outf,O_WRONLY|O_CREAT,0644))<0) {
            perror("open");
            _exit(1);
        }
    } else {
        //pcm_fd = oss_audio_open(NULL, stereo ? 2 : 1, outRate);
        pcm_fd = alsa_audio_open(NULL, stereo ? 2 : 1, outRate);
    }

	TfmxInit();
	StartSong(songnum,0);
	x=0;
	x=30;
	hdb[0]=(struct Hdb){0,0x1C01,0x3200,0x15BE,
		&smplbuf[0x4],&smplbuf[0x4+0x1C42],
		0x40,3,&LoopOff,0,NULL};

    n = 16384 * (stereo ? 2 : 1);
	while (try_to_makeblock(buf));
	while (try_to_makeblock(buf)>=0) {
        if (bqueue) {
            if (toOutFile) {
                write(pcm_fd, buf, n);
            } else {
                alsa_audio_write_s16(pcm_fd, buf, n>>1);
            }
            bqueue--;
        }
	}

    close(pcm_fd);
    free(smplbuf);
	return(0);
}

